package com.fahri.matlab.server;

import com.rabbitmq.client.AMQP.BasicProperties;
import com.rabbitmq.client.Channel;
import com.rabbitmq.client.Connection;
import com.rabbitmq.client.ConnectionFactory;
import com.rabbitmq.client.QueueingConsumer;
import jmatlink.JMatLink;
import org.apache.commons.lang3.ArrayUtils;

import javax.imageio.ImageIO;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.ByteArrayOutputStream;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class MatlabServer {

    static final String QUEUE_NAME = "MATLAB";

    static final String PROMPT = "[matlab-server]";

    public static void main(String[] args) {

        if(args.length != 1 || (args.length == 1 && !isValidIpAddress(args[0]))){
            printHelp();

            System.exit(0);
        }

        String brokerIpAddress = args[0];

        JMatLink engine = new JMatLink();
        Map<String, Long> sessionId = new HashMap<String, Long>();
        Map<String, Byte[]> sessionLastImage = new HashMap<String, Byte[]>();
        Map<String, HashSet<String>> sessionVariableList = new HashMap<String, HashSet<String>>();
        try {

            ConnectionFactory connFactory = new ConnectionFactory();
            connFactory.setHost(brokerIpAddress);
            Connection conn = connFactory.newConnection();
            final Channel ch = conn.createChannel();

            ch.queueDeclare(QUEUE_NAME, false, false, false, null);

            ch.basicQos(1);
            QueueingConsumer consumer = new QueueingConsumer(ch);
            ch.basicConsume(QUEUE_NAME, false, consumer);

            System.out.println(PROMPT + " Awaiting RPC requests");
            while (true) {
                QueueingConsumer.Delivery delivery = consumer.nextDelivery();

                BasicProperties props = delivery.getProperties();
                BasicProperties replyProps = new BasicProperties();
                replyProps.setCorrelationId(props.getCorrelationId());

                String message = new String(delivery.getBody());

                System.out.println(PROMPT + " Received messages : " + message);

                String[] request = message.split("\\|");
                String cmd = request[0];
                String resultQuery = "";
                byte[] resultImage = null;
                String resultClose = "";
                if ("query".equals(cmd)) {
                    String token = request[1];
                    String query = request[2];

                    long id = 0;
                    if (sessionId.containsKey(token)) {
                        id = sessionId.get(token);
                    } else {
                        id = engine.engOpenSingleUse();
                        sessionId.put(token, id);
                    }

                    System.out.println(String.format("%s [%s] Query : %s", PROMPT, token, query));

                    engine.engEvalString(id, query);

                    String var = "";
                    String expression = "";

                    if (query.split("=").length > 1) {
                        var = query.split("=")[0].trim();
                        expression = query.split("=")[1];

                    } else {
                        if (sessionVariableList.get(token) != null && sessionVariableList.get(token).contains(query)) {
                            var = query;
                        } else {
                            var = "ans";
                        }
                        expression = query.split("=")[0];
                    }


                    if (!"".equals(var)) {
                        StringBuilder buffer = new StringBuilder();
                        double[][] arrayResult = engine.engGetArray(id, var);
                        if (arrayResult == null) {
                            resultQuery = "??? Undefined function or variable ???";
                        } else {
                            if (sessionVariableList.containsKey(token)) {
                                Set<String> listVariable = sessionVariableList.get(token);
                                listVariable.add(var);

                            } else {
                                HashSet<String> newListVariable = new HashSet<String>();
                                newListVariable.add(var);
                                sessionVariableList.put(token, newListVariable);
                            }

                            for (int i = 0; i < arrayResult.length; i++) {
                                for (int j = 0; j < arrayResult[0].length; j++) {
                                    double nilai = arrayResult[i][j];

                                    BigDecimal nilaiDec = new BigDecimal(nilai).setScale(4, RoundingMode.HALF_UP);

                                    buffer.append(nilaiDec + " ");
                                }
                                buffer.append("\n");
                            }

                            resultQuery = buffer.toString();
                        }

                    } else {
                        resultQuery = ""; //kosong
                    }

                    Image im = engine.engGetFigure(id, 1, 400, 350);

                    BufferedImage bufferedImage = new BufferedImage(im.getWidth(null), im.getHeight(null), BufferedImage.TYPE_INT_RGB);
                    Graphics bg = bufferedImage.getGraphics();
                    bg.drawImage(im, 0, 0, null);
                    bg.dispose();

                    //convert BufferedImage to byte array
                    ByteArrayOutputStream baos = new ByteArrayOutputStream();
                    ImageIO.write(bufferedImage, "jpg", baos);
                    baos.flush();
                    resultImage = baos.toByteArray();
                    baos.close();

                    if (sessionLastImage.containsKey(token)) {
                        sessionLastImage.remove(token);
                    }

                    Byte[] temp = ArrayUtils.toObject(resultImage);

                    sessionLastImage.put(token, temp);

                } else if ("getimage".equals(cmd)) {
                    String token = request[1];

                    System.out.println(String.format("%s [%s] Get a generated image", PROMPT, token));

                    if (sessionLastImage.containsKey(token)) {
                        Byte[] image = sessionLastImage.get(token);

                        resultImage = ArrayUtils.toPrimitive(image);

                        sessionLastImage.remove(token);

                        System.out.println(String.format("%s [%s] Sending a image", PROMPT, token));
                    }
                } else if ("close".equals(cmd)) {
                    String token = request[1];

                    if (sessionId.containsKey(token)) {
                        long id = sessionId.get(token);
                        engine.engClose(id);
                    }

                    System.out.println(String.format("%s [%s] Close this session", PROMPT, token));
                }

                if ("query".equals(cmd)) {
                    String response = resultQuery;
                    ch.basicPublish("", props.getReplyTo(), replyProps, response.getBytes());

                } else if ("getimage".equals(cmd)) {
                    byte[] response = resultImage;
                    ch.basicPublish("", props.getReplyTo(), replyProps, response);
                } else if ("close".equals(cmd)) {
                    String response = resultClose;
                    ch.basicPublish("", props.getReplyTo(), replyProps, response.getBytes());
                }

                ch.basicAck(delivery.getEnvelope().getDeliveryTag(), false);
            }

        } catch (Exception ex) {
            System.err.println(PROMPT + " Main thread caught exception: " + ex);
            ex.printStackTrace();
            System.exit(1);
        }
    }

    private static boolean isValidIpAddress(String ipAddress) {
        String IPADDRESS_PATTERN =
                "^([01]?\\d\\d?|2[0-4]\\d|25[0-5])\\." +
                        "([01]?\\d\\d?|2[0-4]\\d|25[0-5])\\." +
                        "([01]?\\d\\d?|2[0-4]\\d|25[0-5])\\." +
                        "([01]?\\d\\d?|2[0-4]\\d|25[0-5])$";

        Pattern pattern = Pattern.compile(IPADDRESS_PATTERN);

        Matcher matcher = pattern.matcher(ipAddress);

        return matcher.matches();

    }

    private static void printHelp() {
        System.out.println("Usage: java -jar matlab-server.jar <message_broker_host>");
        System.out.println("Options:");
        System.out.println("\t<message_broker_host>\tIP address of message broker host");
    }
}